import importlib
import inspect

from pydify_sdk.chatflow import DifyChatFlow  # noqa: F401
from pydify_sdk.constants.base import ChatFlowEvent  # noqa: F401
from pydify_sdk.constants.input import AudioType, DocumentType, ImageType, TransferMethod, VideoType  # noqa: F401
from pydify_sdk.wokeflow import DifyWorkFlow  # noqa: F401

schema_module = importlib.import_module("pydify_sdk.schema")
all_schemas = [
    name for name, obj in inspect.getmembers(schema_module, inspect.isclass) if obj.__module__ == schema_module.__name__
]

# 动态导入所有类
for name in all_schemas:
    globals()[name] = getattr(schema_module, name)


__all__ = [
    # sdk
    "DifyWorkFlow",
    "DifyChatFlow",
    # constants
    "ChatFlowEvent",
    "DocumentType",
    "ImageType",
    "VideoType",
    "AudioType",
    "TransferMethod",
] + all_schemas



import json
from typing import Optional, Self

from pydantic import BaseModel, Field, ValidationError, field_validator, model_validator

from .constants.base import ChatFlowEvent, WorkFlowStatus
from .constants.input import AudioType, DocumentType, ImageType, TransferMethod, VideoType


class FileInput(BaseModel):
    type: DocumentType | ImageType | AudioType | VideoType | str = Field(..., description="The type of the file")
    transfer_method: TransferMethod = Field(..., description="The transfer method of the file")
    url: Optional[str] = Field(None, description="The URL of the file")
    upload_file_id: Optional[str] = Field(None, description="The ID of the uploaded file")

    @model_validator(mode="after")
    def check_url_or_upload_file_id(self) -> Self:
        if self.transfer_method == TransferMethod.URL and self.url is None:
            raise ValidationError("url must be provided if transfer_method is remote_url")
        if self.transfer_method == TransferMethod.FILE and self.upload_file_id is None:
            raise ValidationError("upload_file_id must be provided if transfer_method is local_file")
        return self


class AsyncWorkResultResponse(BaseModel):
    id: str = Field(..., description="The ID of the workflow run")
    workflow_id: str = Field(..., description="The ID of the workflow")
    status: WorkFlowStatus = Field(..., description="The status of the workflow run")
    inputs: dict = Field(default={}, description="The input of the workflow run")
    outputs: dict = Field(default={}, description="The output of the workflow run")
    error: Optional[str] = Field(default=None, description="The error of the workflow run")
    total_steps: int = Field(..., description="The total steps of the workflow run")
    total_tokens: int = Field(..., description="The total tokens of the workflow run")
    created_at: int = Field(..., description="The created timestamp of the workflow run")
    finished_at: Optional[int] = Field(..., description="The finished timestamp of the workflow run")
    elapsed_time: float = Field(..., description="The elapsed time of the workflow run")

    @field_validator("inputs", "outputs", mode="before")
    @classmethod
    def check(cls, value: str | dict) -> dict:
        if isinstance(value, str):
            return json.loads(value)
        return value


class WorkFlowResponse(BaseModel):
    task_id: str = Field(..., description="The ID of the workflow run")
    workflow_run_id: str = Field(..., description="The ID of the workflow run")
    data: AsyncWorkResultResponse = Field(..., description="The data of the workflow run")


class ChatFlowRequest(BaseModel):
    query: str = Field(..., description="The query of the chat")
    inputs: Optional[dict] = Field(default={}, description="The input of the chat")
    history: Optional[list] = Field(default=[], description="The history of the chat")


class ChatFlowResponse(BaseModel):
    event: ChatFlowEvent = Field(..., description="The event of the chat")
    conversation_id: str = Field(..., description="The conversation ID of the chat")
    message_id: str = Field(..., description="The message ID of the chat")
    task_id: str = Field(..., description="The task ID of the chat")
    id: str = Field(..., description="ID")
    mode: str = Field(..., description="app mode")
    answer: str = Field(..., description="The answer of the chat")
    metadata: dict = Field(default={}, description="The metadata of the chat")
    created_at: int = Field(..., description="The created timestamp of the chat")


class ChatFlowStreamResponse(BaseModel):
    event: ChatFlowEvent = Field(..., description="The event of the chat")
    conversation_id: str = Field(..., description="The conversation ID of the chat")
    message_id: str = Field(..., description="The message ID of the chat")
    task_id: str = Field(..., description="The task ID of the chat")
    workflow_run_id: str = Field(..., description="The workflow run ID of the chat")
    data: AsyncWorkResultResponse = Field(..., description="The data of the chat")


class ChatSuggestedResponse(BaseModel):
    result: str = Field(..., description="success or fail")
    data: list[str] = Field(..., description="The suggested list of the chat")


class ChatHistory(BaseModel):
    id: str = Field(..., description="message ID")
    conversation_id: str = Field(..., description="conversation ID")
    inputs: Optional[dict] = Field(default=None, description="input parameters")
    query: str = Field(..., description="input text")
    massage_files: Optional[list[dict]] = Field(default=None, description="input file")
    answer: str = Field(..., description="ai response")
    created_at: int = Field(..., description="created time")
    feedback: Optional[dict] = Field(default=None, description="user feedback")


class ChatHistoryResponse(BaseModel):
    limit: int = Field(..., description="The limit of the chat")
    has_more: bool = Field(..., description="Whether has more chat history")
    data: list[ChatHistory] = Field(..., description="The history list of the chat")


class ChatItem(BaseModel):
    id: str = Field(..., description="conversation ID")
    name: str = Field(..., description="conversation name，create by llm")
    inputs: Optional[dict] = Field(default=None, description="input parameters")
    status: str = Field(..., description="conversation status")
    introduction: str = Field(default="", description="chat introduction")
    created_at: int = Field(..., description="created time")
    updated_at: int = Field(..., description="updated time")


class ChatListResponse(BaseModel):
    limit: int = Field(..., description="The limit of the chat")
    has_more: bool = Field(..., description="Whether has more chat history")
    data: list[ChatItem] = Field(..., description="The history list of the chat")



import logging
import logging.config
from types import TracebackType
from typing import Mapping, Optional, TypeAlias, Union

from .config import settings

_SysExcInfoType: TypeAlias = Union[
    tuple[type[BaseException], BaseException, Optional[TracebackType]],
    tuple[None, None, None],
]
_ExcInfoType: TypeAlias = Union[None, bool, _SysExcInfoType, BaseException]
LOGGING_CONF = {
    "version": 1,
    'disable_existing_loggers': False,
    "formatters": {
        "default_fmt": {"format": "[%(asctime)s][%(levelname)s] [%(funcName)s]: %(message)s"},
    },
    "handlers": {
        "console": {
            "level": "INFO",
            "class": "logging.StreamHandler",
            "formatter": "default_fmt",
        },
    },
    "loggers": {
        "": {"level": "INFO", "handlers": ["console"], "propagate": 0},
    },
    "root": {"level": "INFO", "handlers": ["console"]},
}
logging.config.dictConfig(LOGGING_CONF)

logger = logging.getLogger("dify-sdk")


def info(
    msg: object,
    *args: object,
    exc_info: _ExcInfoType = None,
    stack_info: bool = False,
    stacklevel: int = 1,
    extra: Mapping[str, object] | None = None
):
    if not settings.DIFY_LOGGER_ON:
        return
    logger.info(msg, *args, exc_info=exc_info, stack_info=stack_info, stacklevel=stacklevel, extra=extra)



import json
import re
from typing import BinaryIO, Optional, Type, TypeVar

import requests
from pydantic import BaseModel  # pylint: disable=unused-import
from sseclient import SSEClient

from .config import settings
from .constants.base import HttpMethod
from .utils import info

REQUEST_TIME_OUT = 300
ModelType = TypeVar("ModelType", bound="BaseModel")  # pylint: disable=invalid-name


class DifyApiError(Exception):
    def __init__(self, code: int = 0, message: str = "unknown error", **kwargs: dict) -> None:
        super().__init__()
        self.code = code
        self.message = message
        self.kwargs = kwargs

    def __repr__(self) -> str:
        return self.message + f" -> {self.kwargs}"

    def __str__(self) -> str:
        return self.message + f" -> {self.kwargs}"


class DifySDK:
    def __init__(
        self,
        app_key: str,
        *,
        api_url: str = settings.DIFY_API_URL,
        app_name: str = "",
    ):
        self.api_url = api_url
        self.app_key = app_key
        self.app_name = app_name
        self.headers = {
            "Authorization": f"Bearer {self.app_key}",
            "Content-Type": "application/json",
        }
        self.path_comp = re.compile(r":[a-zA-Z\_]+")

    def _complete_data(self, data: dict, user: str, stream: Optional[bool]) -> dict:
        data["user"] = user
        if stream:
            data["response_mode"] = "streaming"
        elif stream is False:
            data["response_mode"] = "blocking"
        return data

    def _handle_error_response(self, resp: requests.Response) -> None:
        if resp.status_code != 200:
            try:
                err = json.loads(resp.text)
                code = err.pop("code", 0)
                message = err.pop("message", "unknown error")
                raise DifyApiError(code=code, message=message, **err)
            except json.JSONDecodeError as e:
                raise DifyApiError(message=f"request failed: {resp.status_code}") from e

    def _parse_stream_data(self, resp: requests.Response) -> dict:
        self._handle_error_response(resp)
        client = SSEClient(resp)  # type: ignore
        event = {}
        try:
            for item in client.events():
                event = json.loads(item.data)
                if event["event"] in {"workflow_finished"}:
                    return event
            raise DifyApiError(message=f"run failed: {event}")
        except json.JSONDecodeError as e:
            raise DifyApiError(message="load data from stream failed") from e

    def _parse_resp_data(self, resp: requests.Response) -> dict:
        self._handle_error_response(resp)
        try:
            resp_json = resp.json()
        except ValueError as e:
            raise DifyApiError(message="return content is not JSON data") from e
        return resp_json

    def _get_request_path(self, path: str, path_params: Optional[dict] = None) -> str:
        request_path = path
        for key in self.path_comp.findall(path):
            if path_params is None or key[1:] not in path_params:
                raise DifyApiError(message=f"missing path parameters: {key}")
            request_path = path.replace(key, str(path_params[key[1:]]))
        return request_path

    def request(  # pylint: disable=too-many-arguments
        self,
        path: str,
        user: str,
        *,
        data: Optional[dict] = None,
        files: Optional[dict] = None,
        path_params: Optional[dict] = None,
        http_method: HttpMethod = HttpMethod.POST,
        stream: Optional[bool] = None,
        model: Optional[Type[ModelType]] = None,
    ) -> dict | ModelType:
        request_path = self._get_request_path(path, path_params)
        url = f"{self.api_url}/{request_path}"
        data = self._complete_data(data or {}, user, stream)
        info(f"[request]{self.app_name} {http_method.value}, path: {request_path}, params: {data}, stream: {stream}")
        if http_method == "POST":
            if stream:
                response = requests.post(
                    url, json=data, headers=self.headers, files=files, stream=True, timeout=REQUEST_TIME_OUT
                )
            else:
                response = requests.post(url, json=data, headers=self.headers, files=files, timeout=REQUEST_TIME_OUT)
        else:
            response = requests.get(url, params=data, headers=self.headers, timeout=REQUEST_TIME_OUT)
        if stream:
            parsed_resp = self._parse_stream_data(response)
        else:
            parsed_resp = self._parse_resp_data(response)
        info(
            f"[response]{self.app_name} {http_method.value}, path: {request_path}, "
            f"params: {data}, stream: {stream}, response: {parsed_resp}"
        )
        if model:
            return model.model_validate(parsed_resp)
        return parsed_resp

    def system_request(
        self,
        path: str,
        *,
        data: Optional[dict] = None,
        path_params: Optional[dict] = None,
        http_method: HttpMethod = HttpMethod.GET,
    ) -> dict:
        request_path = self._get_request_path(path, path_params)
        url = f"{self.api_url}/{request_path}"
        info(f"[request] {self.app_name} {http_method.value}, path: {request_path}, params: {data}")
        if http_method == HttpMethod.POST:
            response = requests.post(url, json=data, headers=self.headers, timeout=REQUEST_TIME_OUT)
        elif http_method == HttpMethod.DELETE:
            response = requests.delete(url, json=data, headers=self.headers, timeout=REQUEST_TIME_OUT)
        elif http_method == HttpMethod.PUT:
            response = requests.put(url, json=data, headers=self.headers, timeout=REQUEST_TIME_OUT)
        else:
            response = requests.get(url, params=data, headers=self.headers, timeout=REQUEST_TIME_OUT)
        parsed_resp = self._parse_resp_data(response)
        info(
            f"[response] {self.app_name} {http_method.value}, path: {request_path}, "
            f"params: {data}, response: {parsed_resp}"
        )
        return parsed_resp

    def upload_file(self, user: str, file: dict[str, BinaryIO | tuple]) -> dict:
        """上传文件"""
        return self.request("files/upload", files=file, user=user)

    def get_app_info(self) -> dict:
        """获取应用信息"""
        return self.system_request("info", http_method=HttpMethod.GET)

    def get_app_parameters(self) -> dict:
        """获取应用参数"""
        return self.system_request("parameters", http_method=HttpMethod.GET)

    def get_app_tools(self) -> dict:
        """获取工具icon"""
        return self.system_request("meta", http_method=HttpMethod.GET)



from typing import Any, Optional

from pydantic import BaseModel

from .base import DifySDK, ModelType
from .constants.base import HttpMethod
from .schema import (
    ChatFlowResponse,
    ChatFlowStreamResponse,
    ChatHistoryResponse,
    ChatListResponse,
    ChatSuggestedResponse,
    FileInput,
)


class DifyChatFlow(DifySDK):
    def _build_data(
        self, query: str, conversation_id: str, data: Optional[dict | ModelType], files: Optional[list[FileInput]]
    ) -> dict:
        return {
            "query": query,
            "conversation_id": conversation_id,
            "inputs": data.model_dump() if isinstance(data, BaseModel) else data,
            "files": files,
        }

    def chat(
        self,
        user: str,
        query: str,
        *,
        conversation_id: str = "",
        data: Optional[dict | ModelType] = None,
        files: Optional[list[FileInput]] = None,
    ) -> ChatFlowStreamResponse:
        request_data = self._build_data(query, conversation_id, data, files)
        response = self.request("chat-messages", user, data=request_data, stream=True, model=ChatFlowStreamResponse)
        return response  # type: ignore

    def sync_chat(
        self,
        user: str,
        query: str,
        *,
        conversation_id: str = "",
        data: Optional[dict | ModelType] = None,
        files: Optional[list[FileInput]] = None,
    ) -> ChatFlowResponse:
        request_data = self._build_data(query, conversation_id, data, files)
        response = self.request("chat-messages", user, data=request_data, model=ChatFlowResponse)
        return response  # type: ignore

    def stop_chat(self, user: str, task_id: str) -> None:
        self.request(
            "chat-messages/run/:task_id/stop",
            user,
            path_params={"task_id": task_id},
            http_method=HttpMethod.POST,
        )

    def feedback_chat(
        self,
        user: str,
        message_id: str,
        rating: str = "",
        content: str = "",
    ) -> None:
        """消息反馈"""
        self.request(
            "chat-messages/run/:task_id/feedback",
            user,
            data={"rating": rating, "content": content},
            path_params={"message_id": message_id},
            http_method=HttpMethod.POST,
        )

    def get_suggested(self, user: str, message_id: str) -> ChatSuggestedResponse:
        """获取下一个问题建议"""
        return self.request(
            "messages/:message_id/suggested",
            user,
            path_params={"message_id": message_id},
            model=ChatSuggestedResponse,
            http_method=HttpMethod.GET,
        )  # type: ignore

    def get_chat_history(
        self, user: str, conversation_id: str, first_id: Optional[str] = None, limit: int = 20
    ) -> ChatHistoryResponse:
        """获取聊天记录"""
        data: dict[str, Any] = {"conversation_id": conversation_id, "limit": limit}
        if first_id:
            data["first_id"] = first_id
        return self.request(
            "messages",
            user,
            data=data,
            model=ChatHistoryResponse,
            http_method=HttpMethod.GET,
        )  # type: ignore

    def get_conversations(
        self, user: str, last_id: Optional[str] = None, limit: int = 20, sort_by: Optional[str] = None
    ) -> ChatListResponse:
        """获取会话列表"""
        data: dict[str, Any] = {"limit": limit}
        if last_id:
            data["last_id"] = last_id
        if sort_by:
            data["sort_by"] = sort_by
        return self.request(
            "conversations",
            user,
            data=data,
            model=ChatListResponse,
            http_method=HttpMethod.GET,
        )  # type: ignore

    def delete_conversation(self, user: str, conversation_id: str) -> None:
        """删除会话"""
        self.request(
            "conversations/:conversation_id",
            user,
            path_params={"conversation_id": conversation_id},
            http_method=HttpMethod.DELETE,
        )

    def delete_chat_history(self, user: str, conversation_id: str) -> None:
        """删除会话记录"""
        self.request(
            "messages/conversation/:conversation_id",
            user,
            path_params={"conversation_id": conversation_id},
            http_method=HttpMethod.DELETE,
        )

    def rename_conversation(self, user: str, conversation_id: str, name: Optional[str], auto: Optional[bool]) -> None:
        """重命名会话"""
        self.request(
            "conversations/:conversation_id/name",
            user,
            data={"name": name, "auto_generate": auto},
            path_params={"conversation_id": conversation_id},
            http_method=HttpMethod.POST,
        )

    def create_annotation(self, question: str, answer: str) -> dict:
        """创建标注"""
        return self.system_request(
            "apps/annotations",
            data={"question": question, "answer": answer},
            http_method=HttpMethod.POST,
        )

    def update_annotation(self, annotation_id: str, question: str, answer: str) -> dict:
        """更新标注"""
        return self.system_request(
            "apps/annotations/:annotation_id",
            data={"question": question, "answer": answer},
            path_params={"annotation_id": annotation_id},
            http_method=HttpMethod.PUT,
        )

    def delete_annotation(self, annotation_id: str) -> dict:
        """删除标注"""
        return self.system_request(
            "apps/annotations/:annotation_id",
            path_params={"annotation_id": annotation_id},
            http_method=HttpMethod.DELETE,
        )



from .const_basic import DocStrEnum


class DocumentType(DocStrEnum):
    TXT = ("TXT", "txt")
    PDF = ("PDF", "pdf")
    MD = ("MD", "md")
    MARKDOWN = ("MARKDOWN", "markdown")
    HTML = ("HTML", "html")
    XLSX = ("XLSX", "xlsx")
    XLS = ("XLS", "xls")
    DOCX = ("DOCX", "docx")
    CSV = ("CSV", "csv")
    EML = ("EML", "eml")
    MSG = ("MSG", "msg")
    PPTX = ("PPTX", "pptx")
    PPT = ("PPT", "ppt")
    XML = ("XML", "xml")
    EPUB = ("EPUB", "epub")


class ImageType(DocStrEnum):
    PNG = ("PNG", "png")
    JPEG = ("JPEG", "jpeg")
    JPG = ("JPG", "jpg")
    GIF = ("GIF", "gif")
    WEBP = ("WEBP", "webp")
    SVG = ("SVG", "svg")


class AudioType(DocStrEnum):
    MP3 = ("MP3", "mp3")
    M4A = ("M4A", "m4a")
    WAV = ("WAV", "wav")
    WEBM = ("WEBM", "webm")
    ARM = ("ARM", "arm")


class VideoType(DocStrEnum):
    MP4 = ("MP4", "mp4")
    MOV = ("MOV", "mov")
    MPEG = ("MPEG", "mpeg")
    MPGA = ("MPGA", "mpga")


class TransferMethod(DocStrEnum):
    URL = ("remote_url", "图片地址")
    FILE = ("local_file", "上传的文件")



from typing import Optional

from pydantic import BaseModel

from .base import DifySDK, ModelType
from .constants.base import HttpMethod
from .schema import AsyncWorkResultResponse, WorkFlowResponse


class DifyWorkFlow(DifySDK):
    def _build_data(self, data: dict | ModelType) -> dict:
        if isinstance(data, BaseModel):
            return {"inputs": data.model_dump()}
        return {"inputs": data}

    def run(self, user: str, data: dict | ModelType) -> WorkFlowResponse:
        data = self._build_data(data)
        response = self.request("workflows/run", user, data=data, stream=True, model=WorkFlowResponse)
        return response  # type: ignore

    def sync_run(self, user: str, data: dict | ModelType) -> WorkFlowResponse:
        data = self._build_data(data)
        response = self.request("workflows/run", user, data=data, model=WorkFlowResponse)
        return response  # type: ignore

    def get_work_result(self, user: str, workflow_run_id: str) -> AsyncWorkResultResponse:
        """获取工作结果"""
        return AsyncWorkResultResponse.model_validate(
            self.request(
                "workflows/run/:workflow_run_id",
                user,
                path_params={"workflow_run_id": workflow_run_id},
                http_method=HttpMethod.GET,
            )
        )

    def stop_work(self, user: str, task_id: str) -> None:
        """停止工作流"""
        self.request(
            "workflows/run/:task_id/stop",
            user,
            path_params={"task_id": task_id},
            http_method=HttpMethod.POST,
        )

    def get_logs(self, page: int = 1, limit: int = 20, *, status: Optional[str] = None) -> dict:
        """获取workflow日志"""
        return self.system_request(
            "workflows/logs", data={"page": page, "limit": limit, "status": status}, http_method=HttpMethod.GET
        )



from enum import IntEnum, StrEnum
from typing import Any, Self


class MetaConst(type):
    def __setattr__(cls, key: Any, value: Any) -> None:  # pragma: no cover
        raise TypeError(f"Constant {key} can't be modified ")


class Constant(metaclass=MetaConst):
    pass


class DocIntEnum(IntEnum):
    def __new__(cls, value: int, doc: str) -> "DocIntEnum":
        self = int.__new__(cls, value)
        self._value_ = value
        self.__doc__ = doc
        return self

    def __str__(self) -> str:
        return self.__doc__  # type:ignore


class DocStrEnum(StrEnum):
    def __new__(cls, value: str, doc: str) -> "DocStrEnum":
        self = str.__new__(cls, value)
        self._value_ = value
        self.__doc__ = doc
        return self

    def __str__(self) -> str:
        return self.__doc__  # type:ignore

    @classmethod
    def from_str(cls, value: str) -> Self:
        for item in cls:
            if item.value == value:
                return item
        raise ValueError(f"{value} is not a valid value for {cls.__name__}")



from .const_basic import DocStrEnum


class HttpMethod(DocStrEnum):
    GET = ("GET", "GET")
    POST = ("POST", "POST")
    DELETE = ("DELETE", "DELETE")
    PUT = ("PUT", "PUT")


class WorkFlowStatus(DocStrEnum):
    RUNNING = ("running", "运行中")
    SUCCEEDED = ("succeeded", "成功")
    FAILED = ("failed", "失败")
    STOPPED = ("stopped", "中止")


class ChatFlowEvent(DocStrEnum):
    MESSAGE = ("message", "LLM 返回文本块事件")
    MESSAGE_FILE = ("message_file", "文件事件，表示有新文件需要展示")
    MESSAGE_END = ("message_end", "消息结束事件，收到此事件则代表流式返回结束")
    TTS_MESSAGE = ("tts_message", "TTS 音频流事件，即：语音合成输出。")
    TTS_MESSAGE_END = ("tts_message_end", "TTS 音频流结束事件")
    MESSAGE_REPLACE = ("message_replace", "消息内容替换事件")
    WORKFLOW_START = ("workflow_start", "开始事件")
    NODE_START = ("node_start", "节点开始事件")
    NODE_FINISHED = ("node_finished", "节点结束事件")
    WORKFLOW_FINISHED = ("workflow_finished", "结束事件")
    ERROR = ("error", "异常事件")
